""" SAX thin-film model functions """

from __future__ import annotations

import jax.numpy as jnp
from typing import Dict
from .._typing import ComplexFloat, ComplexFloat
from ..constants import pi

__all__ = [
    "prop_i",
    "r_complex",
    "r_fresnel_ij",
    "r_fresnel_ji",
    "t_complex",
    "t_fresnel_ij",
    "t_fresnel_ji",
]


def prop_i(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Phase shift acquired as a wave propagates through medium i

    Args:
        wl : wavelength
        ni : refractive index of medium (at wavelength wl)
        di : thickness of layer (same arb. unit as wl)
    """
    return jnp.exp(1j * 2 * pi * params["ni"] / params["wl"] * params["di"])


def r_complex(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Reflection coefficient

    Args:
        t_amp: transmission amplitude
        t_ang: transmission phase
    """
    r_amp = jnp.sqrt((1.0 - params["t_amp"] ** 2))
    r_ang = params["t_ang"] - pi / 2
    return r_amp * jnp.exp(-1j * r_ang)


def r_fresnel_ij(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Normal incidence amplitude reflection from Fresnel's equations

    Args:
        ni : refractive index of the initial medium
        nj : refractive index of the final medium
    """
    return (params["ni"] - params["nj"]) / (params["ni"] + params["nj"])


def r_fresnel_ji(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Normal incidence amplitude reflection from Fresnel's equations

    Args:
        ni : refractive index of the initial medium
        nj : refractive index of the final medium
    """
    return -1 * r_fresnel_ij(params)


def t_complex(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Transmission coefficient (design parameter)

    Args:
        t_amp: transmission amplitude
        t_ang: transmission phase
    """
    return params["t_amp"] * jnp.exp(-1j * params["t_ang"])


def t_fresnel_ij(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Normal incidence amplitude transmission from Fresnel's equations

    Args:
        ni : refractive index of the initial medium
        nj : refractive index of the final medium
    """
    return 2 * params["ni"] / (params["ni"] + params["nj"])


def t_fresnel_ji(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Normal incidence amplitude transmission from Fresnel's equations

    Args:
        ni : refractive index of the initial medium
        nj : refractive index of the final medium
    """
    return (1 - r_fresnel_ij(params) ** 2) / t_fresnel_ij(params)



""" Useful functions for working with SAX. """

from __future__ import annotations

import pickle

from typing import (
    Any,
    Union,
    Tuple,
    Dict,
)
from ._typing import (
    ComplexFloat,
    ModelParams,
    ModelDict,
    Model,
    is_complex_float,
)

__all__ = [
    "copy_params",
    "get_ports",
    "rename_ports",
    "set_params",
    "validate_params",
    "zero",
]


def copy_params(params: ModelParams) -> ModelParams:
    """copy a parameter dictionary

    Args:
        params: the parameter dictionary to copy

    Returns:
        the copied parameter dictionary

    Note:
        this copy function works recursively on all subdictionaries of the params
        dictionary but does NOT copy any non-dictionary values.
    """
    _params = {}
    for k, v in params.items():
        if isinstance(v, dict):
            _params[k] = copy_params(v)
        elif is_complex_float(v):
            _params[k] = v
        else:
            raise ValueError(
                "params dictionary to copy does not have the right type format"
            )
    return _params


def get_ports(model: Model) -> Tuple[str, ...]:
    """get port names of the model

    Args:
        model: the model dictionary to get the port names from
    """
    ports: Dict[str, Any] = {}
    for key in model.funcs:
        p1, p2 = key
        ports[p1] = None
        ports[p2] = None
    return tuple(p for p in ports)


def rename_ports(model: Model, ports: Union[Dict[str, str], Tuple[str]]) -> Model:
    """rename the ports of a model

    Args:
        model: the model dictionary to rename the ports for
        ports: a port mapping (dictionary) with keys the old names and values
            the new names.
    """
    original_ports = get_ports(model)
    assert len(ports) == len(original_ports)
    if not isinstance(ports, dict):
        assert len(ports) == len(set(ports))
        ports = {original_ports[i]: port for i, port in enumerate(ports)}
    funcs: ModelDict = {
        (ports[p1], ports[p2]): model.funcs[p1, p2] for p1, p2 in model.funcs
    }
    new_model = Model(funcs=funcs, params=model.params)
    return new_model


def set_params(
    params: ModelParams, *compnames: str, **kwargs: ComplexFloat
) -> ModelParams:
    """update a parameter dictionary

    add or update the given keyword arguments to each (sub)dictionary of the
    given params dictionary

    Args:
        params: the parameter dictionary to update with the given parameters
        *compnames: the nested component names for which to set the parameters.
            If left out, the given parameters will be applied globally to all
            (sub)components in the dictionary.
        **kwargs: the parameters to update the parameter dictionary with.

    Returns:
        The modified dictionary.

    Note:
        - Even though it's possible to update parameter dictionaries in place,
          this function is convenient to apply certain parameters (e.g.
          wavelength 'wl' or temperature 'T') globally.
        - This operation never updates the given params dictionary inplace.
        - Any non-float keyword arguments will be silently ignored.

    Example:
        This is how to change the wavelength to 1600nm for each component in
        the nested parameter dictionary::

            params = set_params(params, wl=1.6e-6)

        Or to set the temperature for only the direcional coupler named 'dc'
        belonging to the MZI named 'mzi' in the circuit::

            params = set_params(params, "mzi", "dc", T=30.0)
    """
    _params = {}
    if not compnames:
        for k, v in params.items():
            if isinstance(v, dict):
                _params[k] = set_params(v, **kwargs)
            elif is_complex_float(v):
                if k in kwargs and is_complex_float(kwargs[k]):
                    _params[k] = kwargs[k]
                else:
                    _params[k] = v
            else:
                raise ValueError(
                    "params dictionary to set global parameters for "
                    "does not have the right type format"
                )
    else:
        for k, v in params.items():
            if isinstance(v, dict):
                if k == compnames[0]:
                    _params[k] = set_params(v, *compnames[1:], **kwargs)
                else:
                    _params[k] = set_params(v)
            elif is_complex_float(v):
                _params[k] = v
            else:
                raise ValueError(
                    "params dictionary to set global parameters for "
                    "does not have the right type format"
                )
    return _params


def validate_params(params: ModelParams):
    """validate a parameter dictionary

    Args:
        params: the parameter dictionary. This dictionary should be a possibly
            nested dictionary of floats.
    """
    if not params:
        return

    is_dict_dict = all(isinstance(v, dict) for v in params.values())
    if not is_dict_dict:
        for k, v in params.items():
            msg = f"Wrong parameter dictionary format. Should be a (possibly nested) "
            msg += f"dictionary of floats or float arrays. Got: {k}: {v}{type(v)}"
            assert is_complex_float(v), msg
    else:
        for v in params.values():
            assert isinstance(v, dict)
            validate_params(v)


def zero(params: ModelParams) -> float:
    """the zero model function.

    Args:
        params: the model parameters dictionary

    Returns:
        This function always returns zero.
    """
    return 0.0



""" Load data files in OptSim format as SAX model functions

Note:
    This module is intended to load elementary model *functions* from data
    files in OptSim format. Loading full OptSim entities (for example from
    .moml files) is not supported (yet).

The OptSim data format has the following two-line header::

    TransferMatrixFormat5 <SubFormat>
    <num_inputs> <num_outputs> <num_wl> <min_wl> <max_wl>

Older TransferMatrixFormats (like ``TransferMatrixFormat4`` and below) are not
supported by SAX.

``<SubFormat>`` should be one of the following: ``POWER``, ``REAL``, ``REAL_IMAG``, ``AMP_PHASE``,
``AMP_PHASE_RAD``, ``POWER_PHASE``. (Other SubFormats are note supported by SAX)

The second line of the header can be extended with the following configurations:

  - ``<wavelength_units>``: either ``'wavelength_meters'`` or
    ``'wavelength_microns'`` (Other units are not supported by SAX)
  - ``polarization_mode=[0,1,2]``: ``0``: no polarization dependence, ``1``:
    no cross-polarization, ``2``: also cross-polarization
  - ``wavelength_grouping=[0,1,2]``: ``0``, ``1``: no wavelength grouping,
    ``2``: wavelength grouping for easier phase unwrapping
  - ``unwrap_phase=[0,1,2]``: ``0``: no unwrapping (assume unwrapped), ``1``:
    unwrap during interpolation, ``2``: unwrap while reading the file
    (default). (Not supported by SAX)
  - ``frequency_interp=[0,1]``: ``0``: interpolate on wavelength grid, ``1``:
    interpolate on frequency grid (default). (Not supported by SAX)

"""

from __future__ import annotations

import os
import sys
import warnings
import functools

import numpy as np
import pandas as pd
import jax
import jax.numpy as jnp

from typing import Dict, List, Tuple, NamedTuple, Callable
from .._typing import Array
from ..utils import zero

__all__ = [
    "load_optsim_df",
    "optsim_model_function",
    "amplitude_interpolation_with_grouping",
    "phase_interpolation_with_grouping",
]


def load_optsim_df(
    path: str,
    *s_params: str,
    from_cache: bool = True,
    save_cache: bool = True,
    verbose: bool = False,
) -> Tuple[pd.DataFrame, _df_meta]:
    """Load (or partly load) an data file in OptSim format.

    Args:
        path: the path to the data file
        *s_params: the s parameters to load from the data file in OptSim
            format. Each s-parameter should be a string of the following format:
            "{input_port_idx}_{output_port_idx}_{input_mode_str}_{output_mode_str}"
            usually output_port_idx > input_port_idx and input_mode_str in ("te", "tm").
            If not given, all S-matrix elements will be loaded.
        from_cache: if possible, load the requested DataFrame from cache
        save_cache: pickle the resulting DataFrame for easier loading later.
        verbose: print info while loading the data

    Returns:
        the dataframe containing the data for the given s_params and a tuple
        containing some metadata about the data that was loaded.
    """

    path = os.path.abspath(os.path.expanduser(path))
    component, _ = os.path.splitext(os.path.basename(path))

    if s_params:
        _s_params = [str(param) for param in s_params]
        for i, _param in enumerate(_s_params):
            param = [el.lower().strip() for el in _param.split("_") if el]
            if (
                len(param) not in (2, 4)
                or param[2] not in ("te", "tm")
                or param[3] not in ("te", "tm")
            ):
                raise ValueError(
                    "Wrong format for s_params. Given s_params should be strings of "
                    "the following format: '{i_in}_{i_out}_{mode_in}_{mode_out}', "
                    "where mode_in and mode_out should be either 'TE' or 'TM'"
                )
            try:
                param[0] = str(int(param[0]))
                param[1] = str(int(param[1]))
            except ValueError:
                raise ValueError(
                    "Wrong format for s_params. Given s_params should be strings of "
                    "the following format: '{i_in}_{i_out}_{mode_in}_{mode_out}', "
                    "where i_in and i_out should be integer indices."
                )
            if len(param) == 2:
                param.extend(["te", "te"])
            _s_params[i] = "_".join(str(p) for p in param)
        s_params = tuple(_s_params)

    with open(path, "r") as file:
        path = os.path.abspath(os.path.expanduser(path))
        header1 = next(file).strip().split(" ")
        header2 = next(file).strip().split(" ")

    factorformat = ""
    if len(header1) == 2:
        transfer_matrix_format, subformat = header1
    elif len(header1) == 3:
        transfer_matrix_format, subformat, factorformat = header1
    else:
        raise ValueError(
            f"Error parsing '{path}'. " f"Invalid header (first line):\n\n{header1}"
        )

    if not transfer_matrix_format in ("TransferMatrixFormat5",):
        raise ValueError(
            f"Error parsing '{path}'. "
            f"Unsupported TransferMatrixFormat: '{transfer_matrix_format}'."
        )

    if not factorformat.upper() in ("",):
        raise ValueError(
            f"Error parsing '{path}'. " f"FactorFormat '{factorformat}' not supported."
        )

    _num_inputs, _num_outputs, _num_wl, _min_wl, _max_wl, *args_list = header2
    num_inputs = int(_num_inputs)
    num_outputs = int(_num_outputs)
    num_wl = int(_num_wl)
    min_wl = float(_min_wl)
    max_wl = float(_max_wl)

    possible_wavelength_units = (
        "wavelength_meters",
        "wavelength_microns",
        "frequency",
        "wavenumber",
    )
    wavelength_units_list = [
        arg for arg in args_list if arg in possible_wavelength_units
    ]
    if len(wavelength_units_list) > 1:
        raise ValueError(
            f"Error parsing '{path}': too many 'wavelength_units' "
            f"specified: {wavelength_units_list}"
        )
    wavelength_units = (
        "wavelength_meters" if not wavelength_units_list else wavelength_units_list[0]
    )
    if wavelength_units in ("frequency", "wavenumber"):
        raise ValueError(
            f"Error parsing '{path}': wavelength_units '{wavelength_units}' "
            f"not supported (yet)"
        )

    wavelength_factors = {"wavelength_meters": 1.0, "wavelength_microns": 1.0e-6}
    wavelength_factor = wavelength_factors[wavelength_units]

    names = []
    for i in range(num_inputs + num_outputs):
        if subformat in ("AMP_PHASE", "AMP_PHASE_RAD"):
            names.extend([f"A_n_{i}", f"P_n_{i}"])
        elif subformat in ("POWER_PHASE",):
            names.extend([f"A_n_{i}", f"P_n_{i}"])
        elif subformat in ("REAL_IMAG",):
            names.extend([f"R_n_{i}", f"I_n_{i}"])
        elif subformat in ("POWER", "REAL"):
            names.extend([f"R_n_{i}"])
        else:
            raise ValueError(
                f"Error parsing '{path}'. Subformat '{subformat}' not supported."
            )

    args = {
        k: int(v)
        for k, v in (
            arg.split("=") for arg in args_list if arg not in possible_wavelength_units
        )
    }

    # wavelength_grouping: 0, 1: no wavelength grouping, 2: wavelength grouping for easier phase unwrapping
    wavelength_grouping = args.get("wavelength_grouping", 0)
    if wavelength_grouping not in (0, 1, 2):
        raise ValueError(
            f"Error parsing '{path}': "
            f"wavelength_grouping={wavelength_grouping} not in [0, 1, 2]."
        )

    # unwrap_phase: 0: no unwrapping (assume unwrapped), 1: unwrap during interpolation, 2: unwrap while reading the file (default).
    unwrap_phase = args.get("unwrap_phase", 2)
    if unwrap_phase not in (2,):
        warnings.warn(
            f"unwrap_phase={unwrap_phase} not supported. "
            f"Please unwrap the phase manually."
        )

    # frequency_interp: 0: interpolate on wavelength grid, 1: interpolate on frequency grid (default).
    frequency_interp = args.get("frequency_interp", 1)
    if frequency_interp not in (1,):
        warnings.warn(
            f"frequency_interp={frequency_interp} not supported. "
            f"Please interpolate manually."
        )

    # polarization_mode: 0: no polarization dependendence, 1: no cross-polarization, 2: also cross-polarization
    polarization_mode = args.get("polarization_mode", 0)
    num_polarizations = 1
    dfs: Dict[str, List] = {}
    if polarization_mode in (0,):
        names = [f"{name}_te_te" for name in names]
        num_polarizations = 1
        dfs["te"] = []
    elif polarization_mode in (1,):
        names = [f"{name}_tem_tem" for name in names]
        num_polarizations = 2
        dfs["te"] = []
        dfs["tm"] = []
    elif polarization_mode in (2,):
        names_te = [f"{name}_tem_te" for name in names]
        names_tm = [f"{name}_tem_tm" for name in names]
        names = names_te + names_tm
        num_polarizations = 2
        dfs["te"] = []
        dfs["tm"] = []
    else:
        raise ValueError(
            f"Error parsing '{path}': "
            f"polarization_mode={polarization_mode} not in [0, 1, 2]."
        )

    meta = _df_meta(
        component,
        num_inputs,
        num_outputs,
        num_wl,
        min_wl,
        max_wl,
        wavelength_grouping,
        unwrap_phase,
        frequency_interp,
        polarization_mode,
    )
    cached_path = os.path.join(
        os.path.dirname(path),
        "__pycache__",
        "-".join(str(x) for x in meta)
        + ("-ALL" if not s_params else "-")
        + "-".join(str(x) for x in s_params)
        + ".pkl",
    )

    if from_cache and os.path.exists(cached_path):
        df = pd.read_pickle(cached_path).copy()
        if verbose:
            print(
                f"Loaded from cache. "
                f"Memory usage: {df.memory_usage().sum()/(1024**2):.0f}Mb",
                file=sys.stderr,
            )
        return df, meta

    def _handle_df_chunk(df, idx=0, subformat=subformat):
        if idx * num_wl < total_size // num_polarizations:
            input_mode = "te"
        else:
            input_mode = "tm"
            idx -= int(total_size) // int(num_polarizations) // int(num_wl)
        df["wl"] *= wavelength_factor
        df = df.set_index("wl")
        df.columns = [
            name.replace("_n_", f"_{idx}_").replace("_tem", f"_{input_mode}")
            for name in df.columns
        ]
        if s_params:
            df = df[
                [
                    name
                    for name in df.columns
                    if any(f"_{param}" in name for param in s_params)
                ]
            ]

        if subformat in ("POWER_PHASE",):
            c_a = [name for name in df.columns if name.startswith("A_")]
            df[c_a] = np.sqrt(df[c_a])
            subformat = "AMP_PHASE"

        if subformat in ("AMP_PHASE",):
            c_p = [name for name in df.columns if name.startswith("P_")]
            df[c_p] *= np.pi / 180.0
            subformat = "AMP_PHASE_RAD"

        if subformat in ("POWER",):
            c_r = [name for name in df.columns if name.startswith("R_")]
            df[c_r] = np.sqrt(df[c_r])
            subformat = "REAL"

        if subformat in ("REAL",):
            c_r = np.array(
                [name for name in df.columns if name.startswith("R_")], dtype=object
            )
            c_i = np.array(
                [
                    name.replace("R_", "I_")
                    for name in df.columns
                    if name.startswith("R_")
                ],
                dtype=object,
            )
            v_r = df[c_r].values
            v_i = np.zeros_like(v_r)
            v = np.stack([v_r, v_i], -1).reshape(v_r.shape[0], -1)
            c = np.stack([c_r, c_i], 1).ravel()
            wl = df.index
            df = pd.DataFrame(data=v, columns=c)
            df["wl"] = wl
            df = df.set_index("wl")
            subformat = "REAL_IMAG"

        if subformat in ("REAL_IMAG",):
            c_r = [name for name in np.array(df.columns) if name.startswith("R_")]
            c_i = [name for name in np.array(df.columns) if name.startswith("I_")]
            v_r = np.array(df[c_r].values)
            v_i = np.array(df[c_i].values)
            v_a = np.sqrt(v_r ** 2 + v_i ** 2)
            v_p = np.arctan2(v_i, v_r)
            v = np.stack([v_a, v_p], -1).reshape(v_a.shape[0], -1)
            c = [
                name.replace("R_", "A_").replace("I_", "P_")
                for name in np.array(df.columns)
            ]
            wl = df.index
            df = pd.DataFrame(data=v, columns=c)
            df["wl"] = wl
            df = df.set_index("wl")
            subformat = "AMP_PHASE_RAD"

        return df, input_mode

    num_rows = 0
    memory = 0.0
    total_size = int(num_wl * num_polarizations * (num_inputs + num_outputs))
    df_reader = pd.read_csv(
        path,
        skiprows=2,
        sep=" ",
        names=["wl"] + names,
        header=None,
        index_col=False,
        chunksize=num_wl,
    )
    for idx, df in enumerate(df_reader):
        df, input_mode = _handle_df_chunk(df, idx=idx, subformat=subformat)
        num_rows += df.shape[0]
        memory += df.memory_usage().sum() / (1024 ** 2)
        dfs[input_mode].append(df)
        if verbose:
            print(
                f"{num_rows} rows loaded. Memory usage: {memory:.0f}Mb",
                end="\r",
                file=sys.stderr,
            )

    if polarization_mode == 1:
        df = pd.concat(dfs["te"], 1)
    else:
        df = pd.concat([pd.concat(dfs["te"], 1), pd.concat(dfs["tm"], 1)], 1)

    if save_cache:
        os.makedirs(os.path.dirname(cached_path), exist_ok=True)
        df.to_pickle(cached_path)

    return df, meta


def optsim_model_function(
    path: str,
    input_port: int = 0,
    output_port: int = 1,
    input_mode: str = "te",
    output_mode: str = "te",
    from_cache: bool = True,
    save_cache: bool = True,
) -> Callable:
    """generate a model function from a datafile in OptSim format

    Args:
        path: location of the datafile in OptSim format.
        input_port: the index of the input port
        output_port: the index of the output port
        input_mode: the input mode ('TE' or 'TM')
        output_mode: the input mode ('TE' or 'TM')
        from_cache: if possible, load the requested data file from cache
        save_cache: pickle the resulting data for easier loading later.

    Returns:
        a model function which interpolated the given data file
    """
    path = os.path.abspath(os.path.expanduser(str(path)))
    input_port = int(input_port)
    output_port = int(output_port)
    input_mode = str(input_mode).lower()
    output_mode = str(output_mode).lower()
    assert input_mode in ("te", "tm"), "mode0 should be 'TE' or 'TM'"
    assert output_mode in ("te", "tm"), "mode1 should be 'TE' or 'TM'"
    s_param = f"{input_port}_{output_port}_{input_mode}_{output_mode}"
    df, meta = load_optsim_df(
        path,
        from_cache=from_cache,
        save_cache=save_cache
        # path, s_param, from_cache=from_cache, save_cache=save_cache
    )
    if df.values.size == 0:
        return zero
    wls = jnp.array(df.index.values)
    phi = jnp.array(df[f"P_{s_param}"].values.ravel())
    amp = jnp.array(df[f"A_{s_param}"].values.ravel())

    phase_interpolation = (
        phase_interpolation_with_grouping
        if meta.wavelength_grouping == 2
        else jnp.interp
    )

    def optsim_model_func(params):
        assert isinstance(wls, jnp.ndarray)
        assert isinstance(phi, jnp.ndarray)
        wl = jnp.atleast_1d(params["wl"])
        assert isinstance(wl, jnp.ndarray)
        phase = phase_interpolation(wl, wls, phi)
        assert isinstance(phase, jnp.ndarray)
        amplitude = jnp.interp(wl, wls, amp)
        return amplitude * jnp.exp(1j * phase)

    assert isinstance(wls, jnp.ndarray)
    optsim_model_func.__doc__ = f""" SAX model function for {meta.component}

    Connection:
        {input_port}[{input_mode}] -> {output_port}[{output_mode}]

    Args:
        params: parameter dictionary containing at least the key 'wl' for wavelength.

    Note:
        this interpolation is only accurate in the wavelength
        range [{wls[0]*1e6:.3f}μm, {wls[-2]*1e6:.3f}μm) (up to but not including the last wavelength).
        Any extrapolation outside these bounds can yield unexpected results!

    Data:
        the data for this model function can be found here:
        '{path}'
    """

    return optsim_model_func


def _vmap_interpolation(func):
    return functools.wraps(func)(jax.vmap(func, in_axes=(0, None, None), out_axes=0))


@_vmap_interpolation
def amplitude_interpolation_with_grouping(wl: Array, wls: Array, amp: Array) -> Array:
    """Interpolate amplitude where the given wavelengths and amplitudes have wavelength grouping

    Args:
        wl: the wavelength points at which to evaluate the interpolation
        wls: given wavelengths with known amplitude values. Every two
            wavelengths should be paired close together (wavelength grouping
            format) as to enable accurate amplitude interpolation.
        amp: given amplitudes corresponding to the given wavelengths

    Returns:
        amplitude values for the wavelengths to interpolate for

    Note:
        this interpolation is only accurate in the
        range [wls[0], wls[-2]) (wls[-2] not included).
        Any extrapolation outside these bounds can yield unexpected results!
    """
    damp_dwl = (amp[1::2] - amp[::2]) / (wls[1::2] - wls[::2])
    amp = amp[::2]
    wls = wls[::2]
    dwl = (wls[1:] - wls[:-1]).mean(0, keepdims=True)
    t = (jax.lax.nextafter(wl, wl + dwl) - wls) / dwl
    t = jnp.where(jnp.abs(t) < 1, t, 0)
    assert isinstance(t, jnp.ndarray)
    m0 = jnp.where(t > 0, 1.0, 0.0)
    assert isinstance(m0, jnp.ndarray)
    m1 = jnp.where(t < 0, 1.0, 0.0)
    assert isinstance(m1, jnp.ndarray)
    t = (t * m0).sum(0)
    wl0 = (wls * m0).sum(0)
    wl1 = (wls * m1).sum(0)
    amp0 = (amp * m0).sum(0)
    amp1 = (amp * m1).sum(0)
    damp_dwl0 = (damp_dwl * m0).sum(0)
    damp_dwl1 = (damp_dwl * m1).sum(0)
    _amp0 = amp0 - 0.5 * (wl1 - wl0) * (
        damp_dwl0 * (t ** 2 - 2 * t) - damp_dwl1 * t ** 2
    )
    _amp1 = amp1 - 0.5 * (wl1 - wl0) * (
        damp_dwl0 * (t - 1) ** 2 - damp_dwl1 * (t ** 2 - 1)
    )
    _amp = (1 - t) * _amp0 + t * _amp1
    return _amp


@_vmap_interpolation
def phase_interpolation_with_grouping(wl: Array, wls: Array, phi: Array) -> Array:
    """Interpolate phase where the given wavelengths and phases have wavelength grouping

    Args:
        wl: the wavelength points at which to evaluate the interpolation
        wls: given wavelengths with known phase values. Every two wavelengths
            should be paired close together (wavelength grouping format) as to
            enable accurate phase interpolation.
        phis: given phases corresponding to the given wavelengths

    Returns:
        phase values for the wavelengths to interpolate for

    Note:
        this interpolation is only accurate in the
        range [wls[0], wls[-2]) (wls[-2] not included).
        Any extrapolation outside these bounds can yield unexpected results!
    """
    dphi_dwl = (phi[1::2] - phi[::2]) / (wls[1::2] - wls[::2])
    phi = phi[::2]
    wls = wls[::2]
    dwl = (wls[1:] - wls[:-1]).mean(0, keepdims=True)
    t = (wl - wls + dwl * 1e-5) / dwl
    t = jnp.where(jnp.abs(t) < 1, t, 0)
    assert isinstance(t, jnp.ndarray)
    m0 = jnp.where(t > 0, 1.0, 0.0)
    assert isinstance(m0, jnp.ndarray)
    m1 = jnp.where(t < 0, 1.0, 0.0)
    assert isinstance(m1, jnp.ndarray)
    t = (t * m0).sum(0)
    wl0 = (wls * m0).sum(0)
    wl1 = (wls * m1).sum(0)
    phi0 = (phi * m0).sum(0)
    phi1 = (phi * m1).sum(0)
    dphi_dwl0 = (dphi_dwl * m0).sum(0)
    dphi_dwl1 = (dphi_dwl * m1).sum(0)
    _phi0 = phi0 - 0.5 * (wl1 - wl0) * (
        dphi_dwl0 * (t ** 2 - 2 * t) - dphi_dwl1 * t ** 2
    )
    _phi1 = phi1 - 0.5 * (wl1 - wl0) * (
        dphi_dwl0 * (t - 1) ** 2 - dphi_dwl1 * (t ** 2 - 1)
    )
    phi = jnp.arctan2(
        (1 - t) * jnp.sin(_phi0) + t * jnp.sin(_phi1),
        (1 - t) * jnp.cos(_phi0) + t * jnp.cos(_phi1),
    )
    return phi


class _df_meta(NamedTuple):
    """ OptSim Dataframe metadata """

    component: str
    num_inputs: int
    num_outputs: int
    num_wl: int
    min_wl: float
    max_wl: float
    wavelength_grouping: int
    unwrap_phase: int
    frequency_interp: int
    polarization_mode: int



""" SAX thin-film models """

from __future__ import annotations

import jax.numpy as jnp
from ..core import model
from ..funcs.thinfilm import (
    r_fresnel_ij,
    t_fresnel_ij,
    t_fresnel_ji,
    r_fresnel_ji,
    prop_i,
    r_complex,
    t_complex,
    t_complex,
    r_complex,
)


fresnel_mirror_ij = model(
    funcs={
        ("in", "in"): r_fresnel_ij,
        ("in", "out"): t_fresnel_ij,
        ("out", "in"): t_fresnel_ji,
        ("out", "out"): r_fresnel_ji,
    },
    params={
        "ni": 1.0,
        "nj": 1.0,
    },
)
""" fresnel interface """


propagation_i = model(
    funcs={
        ("in", "out"): prop_i,
        ("out", "in"): prop_i,
    },
    params={
        "ni": 1.0,
        "di": 500.0,
        "wl": 532.0,
    },
)
""" propagation phase """


mirror = model(
    funcs={
        ("in", "in"): r_complex,
        ("in", "out"): t_complex,
        ("out", "in"): t_complex,
        ("out", "out"): r_complex,
    },
    params={
        "t_amp": jnp.sqrt(0.5),
        "t_ang": 0.0,
    },
)
""" fresnel mirror """



""" SAX Photonic Integrated Circuit models """

from __future__ import annotations

from ..core import model
from ..funcs.pic import wg_transmission, dc_transmission, dc_coupling


wg = model(
    funcs={
        ("in", "out"): wg_transmission,
        ("out", "in"): wg_transmission,
    },
    params={
        "length": 25e-6,
        "wl": 1.55e-6,
        "wl0": 1.55e-6,
        "neff": 2.34,
        "ng": 3.4,
        "loss": 0.0,
    },
)
""" waveguide model """

dc = model(
    funcs={
        ("p0", "p1"): dc_transmission,
        ("p1", "p0"): dc_transmission,
        ("p2", "p3"): dc_transmission,
        ("p3", "p2"): dc_transmission,
        ("p0", "p2"): dc_coupling,
        ("p2", "p0"): dc_coupling,
        ("p1", "p3"): dc_coupling,
        ("p3", "p1"): dc_coupling,
    },
    params={"coupling": 0.5},
)
""" directional coupler model """



""" SAX model functions """

from __future__ import annotations

from . import pic
from . import optsim
from . import thinfilm

from .pic import wg_transmission, dc_transmission, dc_coupling
from .optsim import (
    optsim_model_function,
    phase_interpolation_with_grouping,
    amplitude_interpolation_with_grouping,
)



""" SAX core """

from __future__ import annotations

from .utils import zero, rename_ports, get_ports, validate_params, copy_params
from typing import (
    Optional,
    Callable,
    Tuple,
    Dict,
)
from ._typing import (
    ModelParams,
    ModelDict,
    Model,
    ComplexFloat,
)


def circuit(
    models: Dict[str, Model],
    connections: Dict[str, str],
    ports: Dict[str, str],
    params: Optional[ModelParams] = None,
) -> Model:
    """circuit factory

    create a (sub)circuit model from a collection of models and connections

    Args:
        models: a dictionary with as keys the model names and values
            the model dictionaries.
        connections: a dictionary where both keys and values are strings of the
            form "modelname:portname"
        ports: a dictionary mapping portnames of the form
            "modelname:portname" to new unique portnames
        params: override default parameters of the components in the circuit.
            Note that only parameters that are different from the params specified
            in the component models need to be specified.

    Returns:
        the circuit model dictionary with the given port names.

    Example:
        A simple mzi can be created as follows::

            import sax
            mzi = sax.circuit(
                models = {
                    "left": sax.models.pic.dc,
                    "top": sax.models.pic.wg,
                    "bottom": sax.models.pic.wg,
                    "right": sax.models.pic.dc,
                },
                connections={
                    "left:p2": "top:in",
                    "left:p1": "bottom:in",
                    "top:out": "right:p3",
                    "bottom:out": "right:p0",
                },
                ports={
                    "left:p3": "in2",
                    "left:p0": "in1",
                    "right:p2": "out2",
                    "right:p1": "out1",
                },
            )
    """

    models, connections, ports = _validate_circuit_parameters(
        models, connections, ports
    )

    for _name, _model in models.items():
        models[_name] = rename_ports(
            _model, {p: f"{_name}:{p}" for p in get_ports(_model)}
        )
        validate_params(models[_name].params)
    modelnames = [[_name] for _name in models]

    model: Model = Model(funcs={}, params={})
    while len(modelnames) > 1:
        for names1, names2 in zip(modelnames[::2], modelnames[1::2]):
            model1 = models.pop(names1[0])
            model2 = models.pop(names2[0])
            model = _combine_models(
                model1,
                model2,
                None if len(names1) > 1 else names1[0],
                None if len(names2) > 1 else names2[0],
            )
            names1.extend(names2)
            for port1, port2 in [(k, v) for k, v in connections.items()]:
                n1, _ = port1.split(":")
                n2, _ = port2.split(":")
                if n1 in names1 and n2 in names1:
                    del connections[port1]
                    model = _interconnect_model(model, port1, port2)
            models[names1[0]] = model
        modelnames = list(reversed(modelnames[::2]))

    if model.funcs:
        model = rename_ports(model, ports)

    if params:
        model = set_model_params(model, params)

    return model


def model(funcs: ModelDict, params: ModelParams, reciprocal: bool = False) -> Model:
    """model factory

    create a model from model functions and a parameter dictionary.

    Args:
        funcs: a dictionary mapping port name combinations onto functions
        params: a parameter dictionary
        reciprocal: if the model is reciprocal, model functions only need to be
            supplied in one port direction to automatically set the reversed
            direction with the same function.
    """
    model = Model(funcs=funcs, params=params)
    if reciprocal:
        ports = get_ports(model)
        for i, p1 in enumerate(ports):
            for p2 in ports[i:]:
                if not (p1, p2) in model.funcs and not (p2, p1) in model.funcs:
                    continue
                if not (p1, p2) in model.funcs:
                    model.funcs[p1, p2] = model.funcs[p2, p1]
                if not (p2, p1) in model.funcs:
                    model.funcs[p2, p1] = model.funcs[p1, p2]
                assert (
                    model.funcs[p1, p2] is model.funcs[p2, p1]
                ), "non-reciprocal model with reciprocal=True"
    return model


def set_model_params(model: Model, params: ModelParams) -> Model:
    """update the model parameters with a given parameter dictionary

    Args:
        model: the model to update the parameters for
        params: the parameter dictionary with updated parameters. Note that
        only parameters that are different from the originals need to be
        specified.

    Returns:
        the model with updated parameters
    """
    return Model(funcs=model.funcs, params=_set_model_params(model.params, params))


def _set_model_params(modelparams: ModelParams, params: ModelParams) -> ModelParams:
    """update the model parameters with a given parameter dictionary

    Args:
        modelparams: the original model parameters to update the parameters for
        params: the parameter dictionary with updated parameters. Note that
        only parameters that are different from the originals need to be
        specified.

    Returns:
        the updated model parameter dictionary
    """
    new_modelparams = {}
    for k, v in modelparams.items():
        if not k in params:
            new_modelparams[k] = v
        elif isinstance(v, dict):
            _v = params[k]
            assert isinstance(_v, dict)
            new_modelparams[k] = _set_model_params(v, _v)
        else:
            new_modelparams[k] = params[k]

    return new_modelparams


def _combine_models(
    model1: Model,
    model2: Model,
    name1: Optional[str] = None,
    name2: Optional[str] = None,
) -> Model:
    """Combine two models into a combined model (without connecting any ports)

    Args:
        model1: the first model dictionary to combine
        model2: the second model dictionary to combine
        name1: the name of the first model (can be None for unnamed models)
        name2: the name of the second model (can be None for unnamed models)
    """
    funcs: ModelDict = {}
    params: ModelParams = {}
    for _model, _name in [(model1, name1), (model2, name2)]:
        for key, value in _model.funcs.items():
            p1, p2 = key
            _params = copy_params(_model.params)
            if value is zero or _name is None:
                funcs[p1, p2] = value
            else:
                funcs[p1, p2] = _NamedParamsFunc(value, _name)
            if _name is None:
                params = {**params, **_params}
            else:
                params = {**params, **{_name: _params}}
    return Model(funcs=funcs, params=params)


def _interconnect_model(model: Model, k: str, l: str) -> Model:
    """interconnect two ports in a given model

    Args:
        model: the component for which to interconnect the given ports
        k: the first port name to connect
        l: the second port name to connect

    Returns:
        the resulting interconnected component, i.e. a component with two ports
        less than the original component.

    Note:
        The interconnect algorithm is based on equation 6 in the paper below::

          Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
          of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.
    """
    funcs: ModelDict = {}
    ports = get_ports(model)
    for i in ports:
        for j in ports:
            mij = model.funcs.get((i, j), zero)
            mik = model.funcs.get((i, k), zero)
            mil = model.funcs.get((i, l), zero)
            mkj = model.funcs.get((k, j), zero)
            mkk = model.funcs.get((k, k), zero)
            mkl = model.funcs.get((k, l), zero)
            mlj = model.funcs.get((l, j), zero)
            mlk = model.funcs.get((l, k), zero)
            mll = model.funcs.get((l, l), zero)
            if (
                (mij is zero)
                and ((mkj is zero) or (mil is zero))
                and ((mlj is zero) or (mik is zero))
                and ((mkj is zero) or (mll is zero) or (mik is zero))
                and ((mlj is zero) or (mkk is zero) or (mil is zero))
            ):
                continue
            funcs[i, j] = _InterconnectedModelFunc(
                i, j, mij, mik, mil, mkj, mkk, mkl, mlj, mlk, mll
            )
    for key in list(funcs.keys()):
        i, j = key
        if i == k or i == l or j == k or j == l:
            del funcs[i, j]
    return Model(funcs=funcs, params=copy_params(model.params))


def _validate_circuit_parameters(
    models: Dict[str, Model], connections: Dict[str, str], ports: Dict[str, str]
) -> Tuple[Dict[str, Model], Dict[str, str], Dict[str, str]]:
    """validate the netlist parameters of a circuit

    Args:
        models: a dictionary with as keys the model names and values
            the model dictionaries.
        connections: a dictionary where both keys and values are strings of the
            form "modelname:portname"
        ports: a dictionary mapping portnames of the form
            "modelname:portname" to new unique portnames

    Returns:
        the validated and possibly slightly modified models, connections and
        ports dictionaries.
    """

    all_ports = set()
    for name, model in models.items():
        _validate_model_dict(name, model)
        for port in get_ports(model):
            all_ports.add(f"{name}:{port}")

    if not isinstance(connections, dict):
        msg = f"Connections should be a str:str dict or a list of length-2 tuples."
        connections, _connections = {}, connections
        connection_ports = set()
        for conn in _connections:
            connections[conn[0]] = conn[1]
            for port in conn:
                msg = f"Duplicate port found in connections: '{port}'"
                assert port not in connection_ports, msg
                connection_ports.add(port)

    connection_ports = set()
    for connection in connections.items():
        for port in connection:
            if port in all_ports:
                all_ports.remove(port)
            msg = f"Connection ports should all be strings. Got: '{port}'"
            assert isinstance(port, str), msg
            msg = f"Connection ports should have format 'modelname:port'. Got: '{port}'"
            assert len(port.split(":")) == 2, msg
            name, _port = port.split(":")
            msg = f"Model '{name}' used in connection "
            msg += f"'{connection[0]}':'{connection[1]}', "
            msg += f"but '{name}' not found in models dictionary."
            assert name in models, msg
            msg = f"Port name '{_port}' not found in model '{name}'. "
            msg += f"Allowed ports for '{name}': {get_ports(models[name])}"
            assert _port in get_ports(models[name]), msg
            msg = f"Duplicate port found in connections: '{port}'"
            assert port not in connection_ports, msg
            connection_ports.add(port)

    output_ports = set()
    for port, output_port in ports.items():
        if port in all_ports:
            all_ports.remove(port)
        msg = f"Ports keys in 'ports' should all be strings. Got: '{port}'"
        assert isinstance(port, str), msg
        msg = f"Port values in 'ports' should all be strings. Got: '{output_port}'"
        assert isinstance(output_port, str), msg
        msg = f"Port keys in 'ports' should have format 'model:port'. Got: '{port}'"
        assert len(port.split(":")) == 2, msg
        msg = f"Port values in 'ports' shouldn't contain a ':'. Got: '{output_port}'"
        assert ":" not in output_port, msg
        msg = f"Duplicate port found in ports or connections: '{port}'"
        assert port not in connection_ports, msg
        name, _port = port.split(":")
        msg = f"Model '{name}' used in output port "
        msg += f"'{port}':'{output_port}', "
        msg += f"but '{name}' not found in models dictionary."
        assert name in models, msg
        msg = f"Port name '{_port}' not found in model '{name}'. "
        msg += f"Allowed ports for '{name}': {get_ports(models[name])}"
        assert _port in get_ports(models[name]), msg
        connection_ports.add(port)
        msg = f"Duplicate port found in output ports: '{output_port}'"
        assert output_port not in output_ports, msg
        output_ports.add(output_port)

    assert not all_ports, f"Unused ports found: {all_ports}"

    return models, connections, ports


def _validate_model_dict(name: str, model: Model):
    assert isinstance(model, Model), f"'{model}' should be a Model tuple"
    ports = get_ports(model)
    assert ports, f"No ports in model {name}"
    for p1 in ports:
        for p2 in ports:
            msg = (
                f"model {name} port combination {p1}->{p2} is no function or callable."
            )
            assert callable(model.funcs.get((p1, p2), zero)), msg


class _InterconnectedModelFunc:
    def __init__(
        self,
        p1: str,
        p2: str,
        mij: Callable,
        mik: Callable,
        mil: Callable,
        mkj: Callable,
        mkk: Callable,
        mkl: Callable,
        mlj: Callable,
        mlk: Callable,
        mll: Callable,
    ):
        self.p1 = p1
        self.p2 = p2
        self.mij = mij
        self.mik = mik
        self.mil = mil
        self.mkj = mkj
        self.mkk = mkk
        self.mkl = mkl
        self.mlj = mlj
        self.mlk = mlk
        self.mll = mll

    def __call__(self, params: ModelParams) -> ComplexFloat:
        """combine the given model functions.

        Note:
            The interconnect algorithm is based on equation 6 in the paper below::

              Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
              of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.
        """
        vij = self.mij(params)
        vik = self.mik(params)
        vil = self.mil(params)
        vkj = self.mkj(params)
        vkk = self.mkk(params)
        vkl = self.mkl(params)
        vlj = self.mlj(params)
        vlk = self.mlk(params)
        vll = self.mll(params)
        return vij + (
            vkj * vil * (1 - vlk)
            + vlj * vik * (1 - vkl)
            + vkj * vll * vik
            + vlj * vkk * vil
        ) / ((1 - vkl) * (1 - vlk) - vkk * vll)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.p1}, {self.p2})"


class _NamedParamsFunc:
    def __init__(self, func: Callable, name: str):
        self.func = func
        self.name = name

    def __call__(self, params: ModelParams) -> ComplexFloat:
        """make a model function look for its model name in the given parameters

        Args:
            func: the original model function acting on a dictionary of parameters
            name: the name of the model
            params: a dictionary for which the keys are the model names and the
                values are the model parameter dictionaries.
        """
        return self.func(params[self.name])

    def __repr__(self):
        return f"<{repr(self.func)[1:-1]}@{self.name}>"



""" SAX Photonic Integrated Circuit model functions """

from __future__ import annotations

import jax.numpy as jnp
from typing import Dict
from .._typing import ComplexFloat, ComplexFloat
from ..constants import pi

__all__ = ["dc_coupling", "dc_transmission", "wg_transmission"]


def dc_coupling(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Directional coupler coupling

    Args:
        coupling: power coupling of the coupler
    """
    return 1j * params["coupling"] ** 0.5


def dc_transmission(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Directional coupler transmission

    Args:
        coupling: power coupling of the coupler (=1-transmission)
    """
    return (1 - params["coupling"]) ** 0.5


def wg_transmission(params: Dict[str, ComplexFloat]) -> ComplexFloat:
    """Waveguide transmission

    Args:
        wl: wavelength
        neff: waveguide effective index
        ng: waveguide group index (used for linear neff dispersion)
        wl0: center wavelength at which neff is defined
        length: [m] wavelength length
        loss: [dB/m] waveguide loss
    """
    neff = params["neff"]
    dwl = params["wl"] - params["wl0"]
    dneff_dwl = (params["ng"] - params["neff"]) / params["wl0"]
    neff = neff - dwl * dneff_dwl
    phase = jnp.exp(jnp.log(2.0 * pi * neff * params["length"]) - jnp.log(params["wl"]))
    return 10 ** (-params["loss"] * params["length"] / 20) * jnp.exp(1j * phase)



""" SAX """

from __future__ import annotations

__author__ = "Floris Laporte"
__version__ = "0.0.10"


from . import nn
from . import core
from . import utils
from . import funcs
from . import models
from . import constants
from . import _typing as typing

from .core import model, circuit, set_model_params
from .utils import (
    set_params,
    rename_ports,
    get_ports,
    copy_params,
    validate_params,
)
